require 'logger'

class ConsoleAuth
  class Unauthorized < Exception; end;
  class Misconfigured < Exception; end;
  class AuthError < Exception; end;

  require "#{File.dirname(__FILE__)}/restriction"
  require 'active_record'

  def self.environment
    if ENV.has_key? 'RACK_ENV'
      ENV['RACK_ENV']
    elsif ENV.has_key? 'RAILS_ENV'
      ENV['RAILS_ENV']
    else
      raise 'No environment specified. Either RACK_ENV or RAILS_ENV must be specified.'
    end
  end

  def self.parse_restrictions(params = {})
    @@restrictions = Array.new
    params[:dir] ||= '/etc/puppetlabs/console-auth/restrictions/**/*.rb'

    Dir[params[:dir]].each do  |restriction_file|
      eval IO.read(restriction_file)
    end
  end

  def self.init
    @@restrictions = Array.new
    @@env = nil
    @@config = Hash.new

    @@logger = Logger.new('/var/log/pe-console-auth/auth.log')
    @@logger.level = Logger::INFO
    @@logger.datetime_format = "%Y-%m-%d %H:%M:%S"
    @@logger.formatter = proc { |level, date, name, msg|
      "#{level} #{date}: #{msg}\n"
    }

    parse_restrictions

    dbconfig_file = if environment == 'test'
                      File.join(File.dirname(__FILE__), '..', 'integration', 'console-auth', 'database.yml')
                    else
                      File.join(File.dirname(__FILE__), '..', 'db', 'config.yml')
                    end
    dbconfig = YAML::load(File.open dbconfig_file)[environment]
    ActiveRecord::Base.configurations['console_auth'] = dbconfig

    require "#{File.dirname(__FILE__)}/../models/user"
    require "#{File.dirname(__FILE__)}/../models/authorized_user"
    require "#{File.dirname(__FILE__)}/../models/role"
    require "#{File.dirname(__FILE__)}/../models/activation"

    config_files = if environment == 'test'
                     ['config.yml','cas_client_config.yml'].map do |config_file|
                       File.join(File.dirname(__FILE__), '..','integration','console-auth',config_file)
                     end
                   else
                     ['config.yml','cas_client_config.yml'].map do |config_file|
                       "/etc/puppetlabs/console-auth/#{config_file}"
                     end
                   end

    config_files.each do |file|
      if File.readable? file
        @@config.rmerge! YAML::load(IO.read(file))
      end
    end
  end

  def self.active?
    File.exists? '/etc/puppetlabs/httpd/console_apps.d/console_auth.conf' and
      File.exists? '/etc/puppetlabs/httpd/console_apps.d/rubycas-server.conf'
  end

  def self.[](index)
    @@restrictions.find { |restriction| restriction.name == index }
  end

  def self.restrict(name, &blk)
    @@restrictions << ConsoleAuth::Restriction.new(name, &blk)
  end

  def self.restrictions
    @@restrictions.map { |restriction| restriction.name }
  end

  def self.unauthenticated_restrictions
    if allow_global_unauthenticated_access?
      [ /\/.*/ ]
    else
      @@restrictions.map do |restriction|
        restriction.name if restriction.has_unauthenticated_method?
      end.compact
    end
  end

  def self.find_restriction(params)
    unless params.has_key?(:url)
      raise ArgumentError, 'The :url parameter is required'
    end

    @@restrictions.find { |restriction| restriction.match? params[:url] }
  end

  def self.access_control_role(params)
    target_url    = params[:url]
    target_method = params[:method]
    user_roles    = Array(params[:roles])

    restriction = @@restrictions.find { |r| r.match? target_url }

    restrict_roles = if restriction.nil?
      user_roles
    else
      unless restriction[target_method].authorized? :roles => user_roles
        raise ConsoleAuth::Unauthorized
      end

      user_roles.empty? ? restriction[target_method].roles : user_roles
    end

    # We don't need to worry about the :admin role since that role
    # automatically gets roles [:read, :write, :admin]
    if restrict_roles.include? :read  and restrict_roles.include? :write
      'READ_WRITE'
    elsif restrict_roles.include? :read
      'READ_ONLY'
    else
      raise ConsoleAuth::Unauthorized
    end
  end

  def self.authorized?(params)
    target_url    = params[:url]
    target_method = params[:method]
    user_roles    = params[:roles]

    @@restrictions.each do |restriction|
      if restriction.match? target_url
        if user_roles.nil?
          return restriction[target_method].allow_unauthenticated?
        else
          return restriction[target_method].authorized? :roles => user_roles
        end
      end
    end

    not user_roles.nil?
  end

  def self.allow_global_unauthenticated_access?
    return false unless get_config.has_key? 'authentication'
    return false unless get_config['authentication'].has_key? 'global_unauthenticated_access'

    get_config['authentication']['global_unauthenticated_access']
  end

  def self.get_config
    @@config
  end

  def self.authorization_defaults
    raise ConsoleAuth::Misconfigured, "The cas_client_config.yml is missing authorization details." unless get_config['authorization']
    get_config['authorization']
  end
  
  def self.authenticator_details(authenticator)
    authenticator_config = authorization_defaults[authenticator]
    raise ConsoleAuth::Misconfigured, "The cas_client_config.yml file is missing details for the [ #{authenticator} ] authenticator." unless authenticator_config
    authenticator_config
  end

  def self.cas_base_url
    if cas_host = get_config['authentication']['cas_host']
      "https://#{cas_host}/cas/"
    else
      nil
    end
  end

  def self.username(env)
    # XXX: These duplicate session values are not optimal and cas_client.rb should be refactored
    #      to not store the CAS server response in the current session.
    #
    #      Also, unless the env['rack.session']['rack.basic.*'] values are needed for backwards
    #      compatibility these values should be replaced with the env['rack.cas.client.*'] values.
    #
    #      Additionally, these values are stored yet again by the ConsoleAuth::Authorization middleware:
    #        env['rack.session']['CONSOLE_AUTH_ACCOUNT_NAME'] = user.username
    #        env['rack.session']['CONSOLE_AUTH_ACCOUNT_ROLE'] = role
    #        env['rack.session']['CONSOLE_AUTH_AUTHENTICATOR'] = user.origin

    # Fetch the user data from the env hash set by the Rack::Cas::Client.valid_session()
    if env.has_key? 'rack.cas.client.user'
      username = env['rack.cas.client.user']
    # Fetch the user data stored in session by the Rack::ConsoleAuth::BasicAuth middleware
    elsif env['rack.session'].has_key? 'rack.basic.user'
      username = env['rack.session']['rack.basic.user']
    # Fetch the memoized CAS response also stored in the session by Rack::Cas::Client.valid_session()
    elsif env['rack.session'].has_key? 'cas'
      username = env['rack.session']['cas']['username_session_key']
    # No username available, something's wrong
    else
      logger.error "No username available."
      raise ConsoleAuth::AuthError, "No username available"
    end
    username
  end

  def self.get_authenticator(env)
    # There is only one place this data can be. If it's not there, Rack::Cas::Client.valid_session was not called.
    if env.has_key? 'rack.cas.client.authenticator'
      authenticator = env['rack.cas.client.authenticator'].split('::').last
    elsif env['rack.session'].has_key? 'rack.basic.user'
      authenticator = 'Basic Auth'
    else
      logger.error "No authenticator available."
      raise ConsoleAuth::AuthError, "No Authenticator."
    end
    authenticator
  end

  def self.get_user(env)
    user = AuthorizedUser.find_by_username(username(env))
    unless user

      # We only need to know the authenticator and its info to assign it to a new AuthorizedUser.
      authenticator = get_authenticator(env)
      case authenticator
      # Convert to internal designation
      when 'SQLEncrypted', 'Basic Auth'
        authenticator = 'local'
      else
        authenticator = authenticator.downcase
      end

      # Fetch the default role for the current authenticator
      role_name = authenticator_details(authenticator)['default_role']
      user = AuthorizedUser.new do |u|
        u.username = username(env)
        u.roles = [Role.find_by_name(role_name)]
        u.status = 'enabled'
        u.origin = authenticator
      end
      user.save!

    end

    user
  end

  def self.logger
    @@logger
  end
end

class ::Hash
  def rmerge!(other_hash)
    merge!(other_hash) do |key, oldval, newval|
      oldval.class == self.class ? oldval.rmerge!(newval) : newval
    end
  end
end
