require 'ostruct'
require 'csv_export'

class GroupReportController < ApplicationController
  unloadable
  helper :compliance

  include CSVExport

  def _show(tab)
    @node_group_tab = tab
    @node_group = NodeGroup.find(params[:id])

    @date = params[:date].present? ? Date.parse(params[:date]) : Time.zone.now.to_date

    node_ids = @node_group.node_ids

    @summary_data = BaselineUtil.summary_for_day(@date, node_ids)
    @unreviewed_changes_on_other_days = BaselineUtil.unreviewed_changes_on_other_days(@date, node_ids)
    @other_days = BaselineUtil.date_changer_options(@date, node_ids)

    @failure_count = ResourceStatus.all(
      :select => 'COUNT(1) AS count',
      :joins => :report,
      :group => 'report_id HAVING MAX(reports.time)',
      :conditions => {
        :failed => true,
        :reports => {
          :node_id => node_ids,
          :kind => 'inspect',
          :time => (@date...@date.tomorrow)
        }
      }
    ).map(&:count).map(&:to_i).sum

    @has_common_changes = Diff.group_against_own_baselines(@node_group, @date, 0, 1).present?

    case tab
    when 'common' then
      redirect_to :action => 'individual', :date => params[:date] and return unless @has_common_changes

      params[:page] ||= 1
      params[:per_page] ||= BaselineDifference.per_page
      total_count = BaselineDifference.self_comparisons.scoped_by_node_id(node_ids).count(
        '1',
        :conditions => { :node_baselines => { :date => @date } },
        :joins => [:baseline_resource, :report_resource],
        :group => :resource_name,
        :having => 'count(distinct node_baselines.node_id) > 1'
      ).count
      @common_changes = WillPaginate::Collection.create(params[:page], params[:per_page], total_count) do |pager|
        pager.replace Diff.group_against_own_baselines(@node_group, @date, pager.offset, pager.per_page)
      end

    when 'failures' then
      redirect_to :action => :common, :date => @date.to_s(:db) and return if @failure_count.zero?

      @failures = Report.send(:with_exclusive_scope) do
        Report.all(
          :select => 'name, node_id, COUNT(resource_statuses.id) AS failures',
          :group  => 'node_id HAVING MAX(reports.time)',
          :joins  => [:resource_statuses, :node],
          :conditions => {
            :node_id => node_ids,
            :kind => 'inspect',
            :time => (@date...@date.tomorrow),
            :resource_statuses => {
              :failed => true,
            },
          }
        )
      end

    when 'individual' then
      @individual_changes = @node_group.nodes.map do |node|
        diff = BaselineDifference.self_comparisons.scoped_by_node_id(node.id).
          find(:all, :conditions => { :node_baselines => { :date => @date } },
               :select => 'baseline_differences.accepted, ' +
               'count(baseline_differences.id) as count',
               :group => :accepted).
          inject(Hash.new(0)) { |hash, cond| hash.merge(cond.accepted => cond.count.to_i) }

        reports = node.reports.all(
          :group      => 'node_id HAVING MAX(time)',
          :conditions => {
            :kind    => 'inspect',
            :time    => (@date...@date.tomorrow),
          }
        )
        node_failed_resources = reports.sum do |r|
          r.resource_statuses.count(:conditions => { :failed => true })
        end

        OpenStruct.new({
          :node_name => node.name,
          :node_id   => node.id,
          :resource_count_failed     => node_failed_resources,
          :resource_count_unreviewed => diff[nil],
          :resource_count_accepted   => diff[true],
          :resource_count_rejected   => diff[false],
        })
      end
    end

    render 'node_group/show'
  end

  def common
    _show('common')
  end

  def individual
    _show('individual')
  end

  def failures
    _show('failures')
  end

  def export
    respond_to do |format|
      format.csv do
        node_ids = NodeGroup.find(params[:id]).node_ids
        date = params[:date].present? ? Date.parse(params[:date]) : Time.zone.now.to_date
        node_baselines = NodeBaseline.self_comparisons.all(
          :conditions => { :node_id => node_ids, :date => date },
          :joins      => :node,
          :order      => 'name',
          :include    => {
            :baseline => {
              :resources => :properties,
            },
            :differences => {
              :baseline_resource => :properties,
              :report_resource => :events,
            },
          }
        )

        data = node_baselines.map do |nb|
          node = nb.node.name
          Diff.collate_properties_by_resource(nb).map do |resource, properties|
            type, title = resource
            properties.map do |prop, vals|
              [
                node,
                node,
                type,
                title,
                prop,
                vals[:baseline] == vals[:inspected],
                vals[:baseline],
                vals[:inspected],
              ]
            end
          end
        end
        data.flatten!(2)
        send_csv :data => data, :filename => "#{date.to_s(:db)}-group-#{params[:id]}.csv"
      end
    end
  end

  def mark_as_reviewed
    state  = params[:state]

    unless ['accepted', 'rejected'].include?(state)
      render :text => "'#{state}' is not an acceptable state for a BaselineDifference", :status => 400
      return
    end

    group    = NodeGroup.find(params[:id])
    date     = Date.parse(params[:date])
    digest   = params[:digest]
    resource = params[:resource]

    node_ids = group.nodes.map(&:id)

    diff_ids = NodeBaseline.self_comparisons.all(
      :joins      => :differences,
      :conditions => {
        :node_id              => node_ids,
        :date                 => date,
        :baseline_differences => {
          :resource_name          => resource,
          :resource_status_digest => digest,
        },
      },
      :select     => '`baseline_differences`.`id` AS baseline_difference_id'
    ).map(&:baseline_difference_id)

    diffs = BaselineDifference.find_all_by_id(diff_ids)

    unless diffs.present?
      raise ActiveRecord::RecordNotFound, "Could not find any BaselineDifferences matching digest '#{digest}'"
    end

    if state == 'accepted'
      BaselineDifference.accept_all_changes(diffs)
    else
      BaselineDifference.reject_all_changes(diffs)
    end

    summary_data = BaselineUtil.summary_for_day(date, node_ids)

    render :text => summary_data.to_json, :content_type => :json, :status => 200
  end

  def show_node
    redirect_to :controller => 'node_report', :action => 'show', :id => params[:node_id], :date => params[:date]
  end

  def show_conflict
    @node_group = NodeGroup.find(params[:id])
    @date       = Date.parse(params[:date])
    @digest     = params[:digest]
    @resource   = params[:resource]

    node_ids = @node_group.nodes.map(&:id)

    node_diffs = BaselineDifference.find(:all,
      :joins => { :node_baseline => :node },
      :conditions =>
      ["`node_baselines`.`node_id` IN (?)
        AND `node_baselines`.`date` = ?
        AND `node_baselines`.`node_id` = `node_baselines`.`reference_node_id`
        AND COALESCE(`baseline_differences`.`resource_status_digest`, '') = ?
        AND `baseline_differences`.`resource_name` = ?",
          node_ids, @date, @digest.to_s, @resource],
      :select => "`baseline_differences`.`id`,
                  `baseline_differences`.`accepted`,
                  `baseline_differences`.`baseline_resource_id`,
                  `baseline_differences`.`report_resource_id`,
                  `baseline_differences`.`decided_at`,
                  `nodes`.`name` AS node_name,
                  `nodes`.`id` AS node_id",
      :order => "node_name"
    )

    @multiple_baseline_states = node_diffs.map do |node_diff|
      baseline_resource = node_diff.baseline_resource
      report_resource   = node_diff.report_resource

      baseline_properties  = baseline_resource ? baseline_resource.properties.inject({}) { |props, prop| props.merge(prop.name => prop.value) }           : {}
      inspected_properties = report_resource   ? report_resource.events.inject({}) { |props, event| props.merge(event.property => event.previous_value) } : {}

      prop_names = baseline_properties.keys | inspected_properties.keys
      properties = {}

      prop_names.each do |prop|
        properties[prop] = {
          :baseline  => baseline_properties[prop],
          :inspected => inspected_properties[prop]
        }
      end

      {
        :name       => node_diff.node_name,
        :node_id    => node_diff.node_id.to_i,
        :diff_id    => node_diff.id,
        :accepted   => node_diff.accepted,
        :can_change => node_diff.can_change?,
        :properties => properties,
      }
    end

    render 'node_group/multiple_baseline_states'
  end

  private
end
