require 'spec_helper'

describe Summarizer do
  before :each do
    Summarizer.stubs(:delay).returns(Summarizer)
  end

  describe "::compare_group_against_node" do
    it "should build the baseline summary for each node in the group" do
      foo_report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), "../fixtures/reports/foo.yaml")))
      bar_report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), "../fixtures/reports/bar.yaml")))
      baz_report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), "../fixtures/reports/baz.yaml")))

      group = NodeGroup.generate!
      group.nodes = [foo_report.node, bar_report.node, baz_report.node]
      group.save!

      NodeBaseline.create_from_report!(baz_report, baz_report.time)

      Summarizer.compare_group_against_node(group, baz_report.node_id)

      NodeBaseline.find_by_node_id(foo_report.node.id).differences.count.should > 0
      NodeBaseline.find_by_node_id(bar_report.node.id).differences.count.should > 0
      NodeBaseline.find_by_node_id(baz_report.node.id).differences.count.should == 0
    end
  end

  describe "::build_baseline_summary" do
    it "should create baseline difference entries for every differing resource" do
      baseline_report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), "../fixtures/reports/baseline.yaml")))
      report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), "../fixtures/reports/inspect.yaml")))

      node_baseline = NodeBaseline.create_from_report!(baseline_report, baseline_report.time)

      Summarizer.build_baseline_summary(report.id, node_baseline)

      diff_resources = node_baseline.differences.map do |diff|
        resource = diff.report_resource || diff.baseline_resource
        [resource.resource_type, resource.title]
      end

      diff_resources.should =~ [
        ['File', '/etc/sudoers'],
        ['File', '/etc/hosts'],
        ['File', '/baseline'],
        ['File', '/inspect'],
      ]
    end

    it "should mark a nodebaseline as diffed" do
      baseline_report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), "../fixtures/reports/baseline.yaml")))
      report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), "../fixtures/reports/inspect.yaml")))

      node_baseline = NodeBaseline.create_from_report!(baseline_report, baseline_report.time)
      node_baseline.diffed.should == false

      Summarizer.build_baseline_summary(report.id, node_baseline)

      node_baseline.diffed.should == true
    end

    it "should not create difference entries if there are no differences" do
      baseline_report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), "../fixtures/reports/baseline.yaml")))

      node_baseline = NodeBaseline.create_from_report!(baseline_report, baseline_report.time)

      Summarizer.build_baseline_summary(baseline_report.id, node_baseline)

      BaselineDifference.count.should == 0
    end

    it "should exclude failed resources from the difference set" do
      initial = Factory(:inspect_report, :host => "test", :time => Time.now - 20.minutes)
      nb = NodeBaseline.create_from_report! initial, initial.time

      report = Factory(:inspect_report, :host => "test", :time => Time.now)
      Factory(:failed_resource, :report => report)

      Summarizer.build_baseline_summary(report.id, nb)

      BaselineDifference.count(:conditions => { :node_baseline_id => nb.id }).should == 0
      # Make sure we didn't leak it elsewhere, either. :)
      BaselineDifference.count.should == 0
    end
  end

  describe "::queue_diff_summarization" do
    it "should start baseline summary generation for every undiffed node baseline" do
      SETTINGS.baseline_day_end = "2400"
      foo_report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), "../fixtures/reports/foo.yaml")))
      bar_report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), "../fixtures/reports/bar.yaml")))
      baz_report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), "../fixtures/reports/baz.yaml")))

      foo_baseline = NodeBaseline.create_from_report!(foo_report, foo_report.time)
      bar_baseline = NodeBaseline.create_from_report!(bar_report, bar_report.time)
      baz_baseline = NodeBaseline.create_from_report!(baz_report, baz_report.time)

      Summarizer.stubs(:delay).returns(Summarizer)

      Summarizer.expects(:build_baseline_summary).with(foo_report.id, foo_baseline)
      Summarizer.expects(:build_baseline_summary).with(bar_report.id, bar_baseline)
      Summarizer.expects(:build_baseline_summary).with(baz_report.id, baz_baseline)

      Summarizer.queue_diff_summarization
    end

    it "should skip nodes that need diffing, but don't have reports yet" do
      SETTINGS.baseline_day_end = "2400"
      foo_report = Report.create_from_yaml(File.read(File.join(File.dirname(__FILE__), "../fixtures/reports/foo.yaml")))
      foo_baseline = NodeBaseline.create_from_report!(foo_report, foo_report.time)

      # ...and one for the day following.
      foo_baseline.create_for_next_day

      Summarizer.stubs(:delay).returns(Summarizer)
      Summarizer.expects(:build_baseline_summary).with(foo_report.id, foo_baseline)
      Summarizer.queue_diff_summarization
    end
  end
end
